code = "code"
reason = "reason"

MALFORMED_HEADER = {code : 3001, reason : "Malformed header"}
ROUTE_DOES_NOT_EXIST = {code : 3002, reason : "Route does not exist"}
CONVERT_FAILED = {code : 1003, reason : "Convert failed"}
INTERNAL_ERROR = {code : 1011, reason : "Internal error"}
INVALID_STATE = {code : 3003, reason : "Invalid state"}


from traceback import print_exception
from sys import exc_info

def capture_trace():
	exc_type, exc_value, exc_traceback = exc_info()
	print_exception(exc_type, exc_value, exc_traceback)

def reverse_dict(dict_):
	return {v : k for k, v in dict_.items()}


from __future__ import print_function
from time import strftime
from inspect import getmodule, stack, currentframe, getframeinfo
from sys import stderr
from sys import stdout

# Formatting
BOLD = "\033[1m"
HIDDEN = "\033[2m"
UNDERLINE = "\033[4m"
BLINK = "\033[5m"
INVERTED = "\033[7m"
HIDDEN = "\033[8m"

# Backgrounds
BACKGROUND_DEFAULT = "\033[49m"
BACKGROUND_BLACK = "\033[40m"
BACKGROUND_RED = "\033[41m"
BACKGROUND_GREEN = "\033[42m"
BACKGROUND_YELLOW = "\033[43m"
BACKGROUND_BLUE = "\033[44m"
BACKGROUND_MAGENTA = "\033[45m"
BACKGROUND_CYAN = "\033[46m"
BACKGROUND_LIGHT_GRAY = "\033[47m"
BACKGROUND_DARK_GRAY = "\033[100m"
BACKGROUND_LIGHT_RED = "\033[101m"
BACKGROUND_LIGHT_GREEN = "\033[102m"
BACKGROUND_LIGHT_YELLOW = "\033[103m"
BACKGROUND_LIGHT_BLUE = "\033[104m"
BACKGROUND_LIGHT_MAGENTA = "\033[105m"
BACKGROUND_LIGHT_CYAN = "\033[106m"
BACKGROUND_WHITE = "\033[107m"

# Colors
DEFAULT = "\033[39m" 
BLACK = "\033[30m"
RED = "\033[31m"
GREEN = "\033[32m"
YELLOW = "\033[33m"
BLUE = "\033[34m"
MAGENTA = "\033[35m"
CYAN = "\033[36m"
LIGHT_GREY = "\033[37m"
DARK_GRAY = "\033[90m"
LIGHT_RED = "\033[91m"
LIGHT_GREEN = "\033[92m"
LIGHT_YELLOW = "\033[93m"
LIGHT_BLUE = "\033[94m"
LIGHT_MAGENTA = "\033[95m"
LIGHT_CYAN = "\033[96m"

CACHED_MODULES = {}

# If the stack becomes too complex to figure out a caller we go through and assume the first valid module is the caller.
# This works reasonably well but isn't 100% accurate and will only happen if the caller is a thread.
def print_out(message, color, file):
	stack_ = stack()
	# Interestingly the if statement below is not executed when excepting KeyboardInterrupts. Weird.
	# To prevent a crash we assume the module's name is 'Unknown'
	module = "Unknown"
	if getmodule(stack_[2][0]) == None:
		for i in stack_[2:]:
			if getmodule(i[0]) != None:
				try:
					module = CACHED_MODULES[i[0].f_code]
				except KeyError:
					module = getmodule(i[0]).__name__
					CACHED_MODULES[i[0].f_code] = module
	else:
		try:
			module = CACHED_MODULES[stack_[2][0].f_code]
		except KeyError:
			module = getmodule(stack_[2][0]).__name__
			CACHED_MODULES[stack_[2][0].f_code] = module
	print("\033[32m%s\033[0m:%i → %s%s\033[0m" % (
		module, stack_[2][0].f_lineno, color, 
		message), file=file)
	file.flush()


def info(message, color=""):
	print_out(message, color, stdout)


def header(message):
	print_out(message, MAGENTA, stdout)


def warning(message):
	print_out(message, YELLOW, stderr)


def error(message):
	print_out(message, RED, stderr)


def success(message):
	print_out(message, GREEN, stdout)


if __name__ == "__main__":
	info("Hi!", color=UNDERLINE+BACKGROUND_GREEN+RED)
	header("This is a header")
	warning("This is a warning")    # > stderr
	error("This is an error")       # > stderr
	success("Great success!")


from .utils import reverse_dict
from . import log as logging

META_ROUTE = "meta"
META_ROUTE_INDEX = 0

BASE_PEER_EXCHANGE_ROUTES = {META_ROUTE_INDEX : META_ROUTE}
BASE_PEER_REVERSE_EXCHANGE_ROUTES = {META_ROUTE : META_ROUTE_INDEX}


def create_exchange_routes(routes):
	exchange_map = {META_ROUTE_INDEX : META_ROUTE}
	exchange_id = 1
	for route in routes:
		if route != META_ROUTE:
			exchange_map[exchange_id] = route
			exchange_id += 1
	return exchange_map


def convert_exchange_routes(routes):
	exchange_routes = {}
	for key in routes:
		if key.isnumeric() and type(routes[key]) is str:
			exchange_routes[int(key)] = routes[key]
		else:
			return None
	return exchange_routes


def assemble_routing(routes):
	reverse_routes = reverse_dict(routes)
	exchange_routes = create_exchange_routes(routes)
	reverse_exchange_routes = reverse_dict(exchange_routes)
	return reverse_routes, exchange_routes, reverse_exchange_routes


async def meta_route(data, handler):
	route_exchange_successful = False
	peer_exchange_routes = None

	if type(data) is dict:
		peer_exchange_routes = convert_exchange_routes(data)
		if peer_exchange_routes != None:
			route_exchange_successful = True

	if route_exchange_successful:
		handler._base.peer_exchange_routes = peer_exchange_routes
		if handler.debug:
			logging.success("Received peer exchange routes: %s" % str(data))
		handler._base.peer_reverse_exchange_routes = \
			reverse_dict(handler._base.peer_exchange_routes)
		
		if hasattr(handler._base, "on_exchange"):	
			await handler._base.on_exchange(handler)
			
		if hasattr(handler, "on_ready"):
			await handler.on_ready()	
	else:
		logging.error("Received invalid exchange routes.")
		await handler.close(reason="Invalid exchange routes")


from asyncio import get_event_loop

from enum import Enum
from abc import ABC, abstractmethod

from websockets import serve, connect
from websockets.framing import OP_BINARY
from websockets.exceptions import ConnectionClosed, InvalidState

# asyncio used logging from the standard library
from . import log as logging
from .utils import capture_trace

from .header import pack_message, parse_metadata
from .header import HeaderLayoutInvalidError
from .header import METADATA_LENGTH

from .routing import assemble_routing, meta_route
from .routing import BASE_PEER_EXCHANGE_ROUTES
from .routing import BASE_PEER_REVERSE_EXCHANGE_ROUTES
from .routing import META_ROUTE

from .close_reasons import *

from .converters import convert


class Shared(ABC):
	def __init__(self, handler_class, debug=False):
		self.handler_class = handler_class
		self.debug = debug

		# {"meta" : meta_route}
		self.routes = {}
		# {meta_route : "meta"}
		self.reverse_routes = {}
		# {0 : "meta"}
		self.exchange_routes = {}
		# {"meta" : 0}
		self.reverse_exchange_routes = {}

		# Add meta route
		self.route(META_ROUTE)(meta_route)


		# Peer routes have not been received yet. As per convention the meta route
		# has to exist and we need it for our first send to succeed (otherwise it
		# would fail during route lookup).
		self.peer_exchange_routes = BASE_PEER_EXCHANGE_ROUTES.copy()
		self.peer_reverse_exchange_routes = \
			BASE_PEER_REVERSE_EXCHANGE_ROUTES.copy()


	@abstractmethod
	def start(self, address, port):
		pass


	def route(self, route):
		def wrapped(func):
			routes = self.routes.copy()
			routes[route] = func
			
			self.routes = routes
			
			self.reverse_routes, \
			self.exchange_routes, \
			self.reverse_exchange_routes = assemble_routing(self.routes)
		return wrapped


	async def consumer(self, websocket):
		handler = self.handler_class(websocket, self)

		await self.on_connect(handler)

		while websocket.open:
			# Receive message
			try:
				message = await websocket.recv()
			except ConnectionClosed as e:
				handler.on_close(e.code, e.reason)
				return
			
			# Extract data type and route id
			try:
				data_type, m_route = parse_metadata(message)
			except HeaderLayoutInvalidError:
				await handler.close(**MALFORMED_HEADER)
				return
			
			# Convert route id back to string
			try:
				route = self.exchange_routes[m_route]
			except KeyError:
				await handler.close(**ROUTE_DOES_NOT_EXIST)
				return
			
			try:
				data = convert(message[METADATA_LENGTH:], data_type)
			except ConvertFailedError:
				await handler.close(**CONVERT_FAILED)
				return

			if self.debug:
				data_repr = str(data).replace("\n", " ")
				if len(data_repr) > 80:
					data_repr = data_repr[:80] + "..."
				logging.info("Received '%s' on route '%s': %s (%s:%d)" % (
					type(data).__name__,
					route, data_repr, websocket.remote_address[0],
					websocket.remote_address[1]))
			
			# Call route
			try:
				await self.routes[route](data, handler)
			except Exception as e:
				await handler.close(**INTERNAL_ERROR)
				capture_trace()
				return


	async def on_connect(self, handler):
		pass


	async def on_exchanged(self, handler):
		pass


class Handler:
	def __init__(self, websocket, base):
		self._websocket = websocket
		self._base = base
	

	@property
	def open(self):
		return self._websocket.open


	@property
	def routes(self):
		return self._base.routes


	@property
	def strict(self):
		return self._base.strict


	@property
	def debug(self):
		return self._base.debug


	@property
	def remote_address(self):
		return self._websocket.remote_address


	async def close(self, code=1000, reason=""):
		self.on_close(code, reason)
		await self._websocket.close(code=code, reason=reason)


	async def send(self, data, route):
		try:
			await self._websocket.write_frame(OP_BINARY, 
				pack_message(
					data,
					self._base.peer_reverse_exchange_routes[route]
				)
			)
		except KeyError:
			await self.close(**ROUTE_DOES_NOT_EXIST)
		except InvalidState:
			await self.close(**INVALID_STATE)
		else:
			if self.debug:
				data_repr = str(data).replace("\n", " ")
				if len(data_repr) > 80:
					data_repr = data_repr[:80] + "..."
				logging.info("Sending '%s' on route '%s': %s (%s:%d)" % (
					type(data).__name__, route, data_repr, 
					self.remote_address[0], self.remote_address[1]))


	async def on_ready(self):
		pass


	def on_close(self, code, reason):
		if code != 1000:
			logging.error("Connection closed (%i): '%s'" % (code, 
				reason))


class Server(Shared):
	def __init__(self, handler_class, debug=False):
		super().__init__(handler_class, debug=debug)


	async def on_connect(self, handler):
		await handler.send(self.exchange_routes, META_ROUTE)


	async def server(self, websocket, path):
		await self.consumer(websocket)


	def start(self, address, port):
		start_server = serve(self.server, address, port)

		get_event_loop().run_until_complete(start_server)
		get_event_loop().run_forever()



class Client(Shared):
	def __init__(self, handler_class, debug=False):
		super().__init__(handler_class, debug=debug)


	async def on_exchange(self, handler):
		await handler.send(handler._base.exchange_routes, META_ROUTE)


	async def client(self, address, port, wss):
		async with connect("ws%s://%s:%i" % ("s" if wss else "", 
			address, port)) as websocket:
			
			await self.consumer(websocket)


	def start(self, address, port, wss=False):
		get_event_loop().run_until_complete(self.client(
			address, port, wss))


from ujson import loads as json_loads

NoneType = None.__class__
CONVERT_COLLECTION = lambda data: json_loads(data.decode())

CONVERTERS = {
	str : lambda data: data.decode(),
	int : lambda data: int(data.decode()),
	float : lambda data: float(data.decode()),
	bool : lambda data: bool(int(data.decode())),
	list : CONVERT_COLLECTION,
	dict : CONVERT_COLLECTION,
	tuple : CONVERT_COLLECTION,
	bytes : lambda data: data,
	NoneType : lambda data: None
}


class ConvertFailedError(ValueError):
	def __init__(self, original_error):
		super(ValueError, self).__init__("conversion failed")


class UnknownDataTypeError(ConvertFailedError):
	def __init__(self, type_):
		super(ConvertFailedError, self).__init__("data type unknown: '%s'" % type_)


def convert(binary_data, data_type):
	try:
		return CONVERTERS[data_type](binary_data)
	except KeyError:
		raise UnknownDataTypeError(data_type)
	except ValueError:
		raise ConvertFailedError()


from .utils import reverse_dict

NoneType = None.__class__

DATA_TYPES = {
	str : 0, 
	int : 1,
	float : 2, 
	bool : 3,
	list : 4, 
	dict : 5,
	tuple : 6,
	bytes : 7,
	NoneType : 8 
}


REVERSE_DATA_TYPES = reverse_dict(DATA_TYPES)


from struct import pack as struct_pack
from struct import unpack as struct_unpack
from struct import error as struct_error
from struct import calcsize as struct_calcsize

from .preparers import prepare

from .data_types import DATA_TYPES, REVERSE_DATA_TYPES

PACK_FORMAT = "BH"
METADATA_LENGTH = struct_calcsize(PACK_FORMAT)


class HeaderLayoutInvalidError(ValueError):
	def __init__(self, original_error):
		self.original_error = original_error
		super(ValueError, self).__init__("header layout invalid")


def create_metadata(data_type, converted_route):
	return struct_pack(PACK_FORMAT, DATA_TYPES[data_type], converted_route)


def pack_message(data, exchange_route):
	data, original_data_type = prepare(data)
	return create_metadata(original_data_type, exchange_route) + data


def parse_metadata(message):
	try:
		metadata = struct_unpack(PACK_FORMAT, message[:METADATA_LENGTH])
		return REVERSE_DATA_TYPES[metadata[0]], metadata[1]
	except struct_error as e:
		raise HeaderLayoutInvalidError(e)


from ujson import dumps as json_dumps

PREPARE_COLLECTION = lambda data: json_dumps(data).encode()

NoneType = None.__class__

PREPARERS = {
	str : lambda data: data.encode(),
	int : lambda data: str(data).encode(),
	float : lambda data: str(data).encode(),
	# Faster than int(data)
	bool : lambda data: str(1 if data else 0).encode(),
	list : PREPARE_COLLECTION,
	dict : PREPARE_COLLECTION,
	tuple : PREPARE_COLLECTION,
	bytes : lambda data: data,
	NoneType : lambda data: "".encode()
}


class PrepareFailedError(ValueError):
	def __init__(self):
		super(ValueError, self).__init__("preparation failed (data type unsupported)")



def prepare(original_data, encoder=None):
	original_data_type = type(original_data)

	try:
		return PREPARERS[original_data_type](original_data), \
		original_data_type

	except KeyError:
		raise PrepareFailedError()
	except ValueError:
		raise PrepareFailedError()
	except TypeError:
		raise PrepareFailedError()


import os
from setuptools import setup, find_packages
from sys import version_info


description = "Asynchronous, route based, and data type preserving network protocol."
long_description = """
============
highway.py
============

.. image:: https://cloud.githubusercontent.com/assets/9287847/26756438/93c1ff54-48a2-11e7-981e-37aeb7b2383c.png
   :height: 150px
   :alt: highway banner
   :align: center


**highway** is a *lightweight*, *route-based* and *data-type
preserving* network protocol framework built on top of *WebSockets*. It facilitates routes as a means of data transmission.
"""

setup(
	name="highway.py",
	version="0.2.2",
	author="Philip Trauner",
	author_email="philip.trauner@arztpraxis.io",
	url="https://github.com/PhilipTrauner/highway.py",
	packages=find_packages(),
	description=description,
	long_description=long_description,
	install_requires=[
		"websockets==3.4",
		"ujson==1.35"
	],
	license="MIT",
	classifiers=[
		"Development Status :: 4 - Beta",
		"Programming Language :: Python :: 3.6",
		"Intended Audience :: Developers",
		"License :: OSI Approved :: MIT License",
	],
	keywords=["networking", "asynchronous", "data type preserving"]
)
